@@ -11,55 +11,72 @@ import yaml
 import jinja2
 import argparse
 import traceback
+import pathlib
 
 baseDir = os.path.dirname(os.path.abspath(__file__))
 
+if not os.path.exists(f'{baseDir}/version'):
+  print(f"Jinny's version file doesn't exist, expected to find it at {basedir}/version. Not a good sign, might be best to reinstall jinny!")
+  __version__ = "mystery.version"
+
 with open(f'{baseDir}/version') as f:
-  __version__ = f.read()
+  __version__ = f.read().strip("\n")
 
 ##########################################
 # Variables
 
+class LoggingSettings():
+  def __init__(self, location:str="/dev/stdout", colour:bool=True, verbosity:int=0):
+    self.location = location
+    self.colour = colour
+    self.verbosity = verbosity
+
+    self.beginChar = "\033[92m" if colour else ''
+    self.badChar = "\033[91m" if colour else ''
+    self.warnChar = "\033[93m" if colour else ''
+    self.endChar = "\033[0m" if colour else ''
+
+
 globalAllTemplatesProcessed = {}
 baseJ2Env = jinja2.Environment(trim_blocks=True, lstrip_blocks=True, keep_trailing_newline=True)
 CombineLists = False
-VerboseSetting = 0
-LoggingLocation = "/dev/stdout"
+args = None
+
+CurrentLoggingSettings = LoggingSettings()
 
 ##########################################
 # Machinery
 def CombineValues(originalVals, newVals, sourceName:str):
   originalType = type(originalVals)
   newType = type(newVals)
-  if VerboseSetting > 0:
-    print(VerboseSetting)
+  if CurrentLoggingSettings.verbosity > 0:
     Log(f'CombineValues(): handling {sourceName}, original item = {originalType!s}, new item = {newType!s}')
 
   # Easy wins
   if originalType == list and newType == list:
-    if VerboseSetting > 0:
+    if CurrentLoggingSettings.verbosity > 0:
       Log(f'CombineValues(): found two lists, {"combining and returning" if CombineLists else "replacing old list with new one"}')
     return originalVals + newVals if CombineLists else newVals
 
   if originalType == None:
-    if VerboseSetting > 0:
+    if CurrentLoggingSettings.verbosity > 0:
       Log(f'CombineValues(): original item is None, replacing with new item')
     return newVals
   
   if newType == None:
-    if VerboseSetting > 0:
+    if CurrentLoggingSettings.verbosity > 0:
       Log(f'CombineValues(): new item is None, returning None')
     return None
 
   # More Common
   if originalType == dict and newType == dict:
-    if VerboseSetting > 1:
+    if CurrentLoggingSettings.verbosity > 1:
       Log(f'CombineValues(): Handling a dict merge')
     workingVals = originalVals.copy()
     origKeys = originalVals.keys()
     for eleKey, eleVal in newVals.items():
       if eleKey not in origKeys:
-        if VerboseSetting > 1:
+        if CurrentLoggingSettings.verbosity > 1:
           Log(f'CombineValues(): => Adding new key "{eleKey}"')
         workingVals[eleKey] = eleVal
         continue
@@ -67,28 +84,28 @@ def CombineValues(originalVals, newVals, sourceName:str):
       newType = type(eleVal)
       oldType = type(originalVals[eleKey])
       if newType == str or newType == int or newType == bool or newType == float or newType == complex:
-        if VerboseSetting > 1:
+        if CurrentLoggingSettings.verbosity > 1:
           Log(f'CombineValues(): => Updated value for "{eleKey}"')
         workingVals[eleKey] = eleVal
         continue
       if oldType == list and newType == list:
         if originalVals[eleKey] == newVals[eleKey]:
-          if VerboseSetting > 1:
+          if CurrentLoggingSettings.verbosity > 1:
             Log(f'CombineValues(): => Key "{eleKey}" is the same list in both old and new, continuing')
         else:
-          if VerboseSetting > 1:
+          if CurrentLoggingSettings.verbosity > 1:
             Log(f'CombineValues(): => Key "{eleKey}" is an altered list, {"combining and continuing" if CombineLists else "replacing old list with new one"}')
           workingVals[eleKey] = originalVals[eleKey] + newVals[eleKey] if CombineLists else newVals[eleKey]
         continue
 
       if oldType == dict and newType != dict:
-        if VerboseSetting > 1:
+        if CurrentLoggingSettings.verbosity > 1:
           Log(f'CombineValues(): => Key "{eleKey}" was a dict but updated to be a {newType!s}, replacing')
         workingVals[eleKey] = newVals[eleKey]
         continue
 
       if newType == dict and newType == dict:
-        if VerboseSetting > 1:
+        if CurrentLoggingSettings.verbosity > 1:
           Log(f'CombineValues(): => Key "{eleKey}" is a dict updated with another dict, recursively calling CombineValues to handle')
         res = CombineValues(originalVals[eleKey], newVals[eleKey], sourceName)
         workingVals[eleKey] = res
@@ -96,7 +113,7 @@ def CombineValues(originalVals, newVals, sourceName:str):
     return workingVals
 
 def SetNestedValue(baseResource, path:list, value):
-  if VerboseSetting > 1:
+  if CurrentLoggingSettings.verbosity > 1:
     Log(f'SetNestedValue(): => Handling path {".".join(path)} of {len(path)} items')
   currentTier = baseResource
   for index, element in enumerate(path):
@@ -116,9 +133,9 @@ def SetNestedValue(baseResource, path:list, value):
         location = int(element)
       except:
         execDetails = sys.exc_info()
-        Log(f"Found a list at location '{'.'.join(path[:index])}' but did not find a number integer in the targeting, ie failed to convert the next element in the path '{element}' to an integer so can't target the list, details:\nType:{execDetails[0]}\nValue:{execDetails[1]}\nTrace:\n{execDetails[2]}", quitWithStatus=1)
+        Log(f"SetNestedValue(): Found a list at location '{'.'.join(path[:index])}' but did not find a number integer in the targeting, ie failed to convert the next element in the path '{element}' to an integer so can't target the list, details:\nType:{execDetails[0]}\nValue:{execDetails[1]}\nTrace:\n{execDetails[2]}", quitWithStatus=1)
       if location + 1 > len(currentTier):
-        Log(f"Found a list at location '{'.'.join(path[:index])}' that is {len(currentTier)} items long, however '{location}' is not an existing element, hence can't be targeted! Details:\nType:{execDetails[0]}\nValue:{execDetails[1]}\nTrace:\n{execDetails[2]}", quitWithStatus=1)
+        Log(f"SetNestedValue(): Found a list at location '{'.'.join(path[:index])}' that is {len(currentTier)} items long, however '{location}' is not an existing element, hence can't be targeted! Details:\nType:{execDetails[0]}\nValue:{execDetails[1]}\nTrace:\n{execDetails[2]}", quitWithStatus=1)
       Log(f'SetNestedValue(): => Moving to nested list property index "{location}"')
       currentTier = currentTier[location]
     elif currentType == str or currentType == int or currentType == bool or currentType == float or currentType == complex:
@@ -126,7 +143,7 @@ def SetNestedValue(baseResource, path:list, value):
         raise Exception(f"Hit a non-nested type of {currentType} at {index} / {len(path)} after cycling through {', '.join(path[:index])}.\nFull path is {', '.join(path)}.\nTarget Object is:\n{yaml.dump(baseResource)}".encode().decode('unicode-escape'))
 
 def GenerateNestedDict(path:list, value):
-  if VerboseSetting > 1:
+  if CurrentLoggingSettings.verbosity > 1:
     Log(f'GenerateNestedDict(): => Generating a nested dict {len(path) - 1} levels deep setting "{path[len(path)-1]}" to "{value}"')
   replicatedObj = {}
   target = replicatedObj
@@ -139,18 +156,69 @@ def GenerateNestedDict(path:list, value):
   return replicatedObj
 
 
+def ParseValuesFile(path:str):
+  ext = os.path.basename(path).split(".")
+  if len(ext) == 2:
+    ext = ext[1]
+  else:
+    ext = None
+  finalObj = None
+  with open(path) as f:
+    fileData = f.read()
+
+  if ext == "yaml" or ext == "yml":
+    try:
+      ymlObj = yaml.load(fileData, Loader=yaml.FullLoader)
+      return ymlObj
+    except Exception as e:
+      Log(f"Main(): Inputs file at path '{path}' has a YAML extension naming convention but fails YAML parsing.", AlwaysLog=True)
+      if CurrentLoggingSettings.verbosity > 1:
+        Log(f'Main(): Full stack trace:\n\n{traceback.format_exc()}', AlwaysLog=True)
+      Log(f"Main(): Error:\n{e.problem}\n{e.problem_mark}", AlwaysLog=True)
+      Log("Main(): Exiting!", quitWithStatus=2)
+
+  elif ext == "json":
+    try:
+      jsonObj = json.loads(fileData)
+      return jsonObj
+    except Exception as e:
+      Log(f"Main(): Inputs file at path '{path}' has a JSON extension naming convention but fails JSON parsing.", AlwaysLog=True)
+      if CurrentLoggingSettings.verbosity > 1:
+        Log(f'Main(): Full stack trace:\n\n{traceback.format_exc()}', AlwaysLog=True)
+      Log(f"Main(): Error: {getattr(e, 'msg')} at line {getattr(e, 'lineno')} column {e.colno}", AlwaysLog=True)
+      Log("Main(): Exiting!", quitWithStatus=2)
+
+  else:
+    try:
+      ymlObj = yaml.load(fileData, Loader=yaml.FullLoader)
+      return ymlObj
+    except Exception as e:
+      pass
+    
+    try:
+      jsonObj = json.loads(fileData)
+      return jsonObj
+    except Exception as e:
+      pass
+    Log(f"Main(): Could not load inputs file '{path}' as either a JSON or a YAML file, please inspect!", quitWithStatus=1)
+
 ##########################################
 # Logging
-def Log(message, quit:bool=False, quitWithStatus:int=1, AlwaysVerbose:bool=True):
-  if AlwaysVerbose == False and VerboseSetting == False:
+def Log(message, quit:bool=False, quitWithStatus:int=0, AlwaysLog:bool=False):
+  if AlwaysLog == False and CurrentLoggingSettings.verbosity == 0 and not quit and quitWithStatus == 0:
     return
-  if type(message) is dict:
-    logToFile(LoggingLocation, f'<{TimeStamp()}> - ' + "\n => ".join([k + ": " + str(message[k]) for k in message]))
-  else:
-    logToFile(LoggingLocation, f'<{TimeStamp()}> - {message}')
 
-  if quit or quitWithStatus > 1:
-    exit(1 * quitWithStatus)
+  if quit or quitWithStatus > 0:
+    if type(message) is dict:
+      logToFile(CurrentLoggingSettings.location, f'{CurrentLoggingSettings.badChar}\n*********************\n<{TimeStamp()}>{CurrentLoggingSettings.endChar} - ' + "\n => ".join([k + ": " + str(message[k]) for k in message]))
+    else:
+      logToFile(CurrentLoggingSettings.location, f'{CurrentLoggingSettings.badChar}\n*********************\n<{TimeStamp()}>{CurrentLoggingSettings.endChar} - {message}')
+    exit(max(1, quitWithStatus))
+
+  if type(message) is dict:
+    logToFile(CurrentLoggingSettings.location, f'{CurrentLoggingSettings.warnChar if AlwaysLog else CurrentLoggingSettings.beginChar}<{TimeStamp()}>{CurrentLoggingSettings.endChar} - ' + "\n => ".join([k + ": " + str(message[k]) for k in message]))
+  else:
+    logToFile(CurrentLoggingSettings.location, f'{CurrentLoggingSettings.warnChar if AlwaysLog else CurrentLoggingSettings.beginChar}<{TimeStamp()}>{CurrentLoggingSettings.endChar} - {message}')
 
 ##############################################
 # Internal Functions
@@ -161,13 +229,15 @@ def logToFile(path:str, msg:str):
   if path == "/dev/stdout":
     sys.stdout.write(msg + "\n")
   else:
-    with open(path, 'a') as f:
+    with open(path, 'a+') as f:
       f.write(msg + "\n")
 
 ##########################################
 # Argparsing
 def ArgParsing():
-  parser = argparse.ArgumentParser(description='''Jinny handles complext templating for jinja templates at a large scale and with multiple inputs and with a decent amount of customisation available.
+  global CurrentLoggingSettings
+  parser = argparse.ArgumentParser(description=f'''jinny v{__version__} | jinny.scripted.dog
+Jinny handles complex templating for jinja templates at a large scale and with multiple inputs and with a decent amount of customisation available.
 
 Commonly you'll want to utilse very straight forward features, such as:
 
@@ -180,8 +250,11 @@ $ jinny -t template.yml -i base-values.yml overrides.json
 => Add even more overrides via environment variables, so your pipelines can completely replace any bad value:
 $ JINNY_overridden_value="top-priority" jinny -t template.yml -i base-values.yml overrides.json
 
+=> Or via CLI:
+$ jinny -t template.yml -i base-values.yml -e overridden_value="top-priority" overrides.json
+
 => Pump all your files to a single stdout stream with a separator so different files are clearly marked:
-$ jinny -t template-1.yml template-2.yml -i inputs.json -s '---'
+$ jinny -t template-1.yml template-2.yml -i inputs.json -s='---'
 
 => Dump all your templated files into a directory for capture
 $ jinny -t template-1.yml template-2.yml -i inputes.json -d /path/to/directory
@@ -193,12 +266,13 @@ $ jinny -t template-1.yml -i inputs.json | kubectl apply -f -
 
 You can modify jinja's environment settings via the rest of the command line options. Please note that jinny is opinionated and automatically strips line space from templates. You can, of course, turn this off!
 
-''')
+''', formatter_class=argparse.RawTextHelpFormatter)
 
   # Core arguments
   parser.add_argument("-v", "--verbose", help="Set output to verbose", action="store_true")
   parser.add_argument("-vvv", "--super-verbose", help="Set output to super verbose where this script will print basically everything", action="store_true")
-  parser.add_argument("-i", "--input-values", help="Add one or more file locations that include input values to the templating", action="append", nargs="*")
+  parser.add_argument("-i", "--inputs", help="Add one or more file locations that include input values to the templating", action="append", nargs="*")
+  parser.add_argument("-e", "--explicit", help="Explicitly define a variable that trumps all other variables using a variable=value format. Adding variables like this trumps every other setting for that variable", action="append", nargs="*")
   parser.add_argument("-t", "--templates", help="Add one or more file locations that contain the templates themselves", action="append", nargs="*", required=True)
   parser.add_argument("-ie", "--ignore-env-vars", help="Tell jinny to ignore any environment variables that begin with JINNY_, defaults to not ignoring these environment variables and setting them at the highest priority", action="store_true")
   parser.add_argument("-ds", "--dict-separator", help="When providing targeting on the CLI or via environment variables, choose a particular separating character for targeting nested elements, defaults to '.'", default=".", type=str)
@@ -220,18 +294,14 @@ You can modify jinja's environment settings via the rest of the command line opt
   parser.add_argument("-d", "--dump-to-dir", help="Dump completed templates to a target directory", type=str)
   parser.add_argument("-s", "--stdout-seperator", help="Place a seperator on it's own individual new line between successfully templated template when printing to stdout, eg '---' for yaml", type=str, default='')
   parser.add_argument("-c", "--combine-lists", help="When cascading values across multiple files and encountering two lists with the same key, choose to combine the old list with the new list rather than have the new list replace the old", action="store_false")
+  parser.add_argument("-ld", "--log-destination", help="Chose an alternate destination to log to, jinny defaults to stdout but you can provide a file to print output to instead", default="/dev/stdout", type=str)
+  parser.add_argument("-nc", "--no-color", "--no-colour", help="Turn off coloured output", action="store_false")
 
   args = parser.parse_args()
 
   if args.combine_lists:
     CombineLists = True
 
-  if args.verbose:
-    VerboseSetting = 1
-
-  if args.super_verbose:
-    VerboseSetting = 2
-
   # jinja2.Environment(bl)
   baseJ2Env = jinja2.Environment(
     block_start_string = args.j_block_start,
@@ -246,6 +316,28 @@ You can modify jinja's environment settings via the rest of the command line opt
     keep_trailing_newline = args.j_keep_trailing_newline
   )
 
+  # Checking that the directory for the logging location exists
+  if args.log_destination != "/dev/stdout" and not os.path.isdir(os.path.dirname(args.log_destination)):
+    Log(f"ArgParsing(): The parent directory for the custom log file '{args.log_destination}' either does not exist or is not a directory, please check it!", quitWithStatus=1)
+
+  if type(args.dump_to_dir) == str:
+    args.dump_to_dir = args.dump_to_dir.strip('/')
+    if not os.path.isdir(os.path.dirname(args.dump_to_dir)):
+      Log(f"ArgParsing(): The parent directory for dumping completed templates to '{args.dump_to_dir}' does not exist, please check your arguments!", quitWithStatus=1)
+
+    if os.path.isfile(args.dump_to_dir):
+      Log(f"ArgParsing(): The location provided for dumping completed templates to is a file and not a directory '{args.dump_to_dir}', please check your arguments!", quitWithStatus=1)
+      
+    if not os.path.exists(args.dump_to_dir):
+      os.mkdir(args.dump_to_dir)
+
+  vs = 0
+  if args.verbose:
+    vs = 1
+  if args.super_verbose:
+    vs = 2
+  CurrentLoggingSettings = LoggingSettings(location = args.log_destination, colour = args.no_color, verbosity = vs)
+
   return args
 
 ##########################################
@@ -262,13 +354,13 @@ class TemplateHandler():
     if path != "":
       self.name = os.path.abspath(path)
       if not os.path.exists(path):
-        Log(f"Failed to read template at path '{path}', cannot load in desired template!", quitWithStatus=1)
+        Log(f"TemplateHandler(): Failed to read template at path '{path}', cannot load in desired template!", quitWithStatus=1)
       with open(path, "r") as f:
         self.templateData = f.read()
         self.basename = os.path.basename(self.name)
     else:
       if not templateName:
-        raise Exception(f'A template name must be provided if passing a raw string template!')
+        raise Exception(f'TemplateHandler(): A template name must be provided if passing a raw string template!')
       self.name = "raw provided string"
       self.templateData = rawString
       self.basename = templateName
@@ -277,7 +369,7 @@ class TemplateHandler():
       self.loadedTemplate = baseJ2Env.from_string(self.templateData)
     except Exception as e:
       execDetails = sys.exc_info()
-      Log(f"Failed to load template at '{path}' with an exception from Jinja, details:\nType:{execDetails[0]}\nValue:{execDetails[1]}\nTrace:\n{traceback.format_exc()}", quitWithStatus=1)
+      Log(f"TemplateHandler(): Failed to load template at '{path}' with an exception from Jinja, details:\nType:{execDetails[0]}\nValue:{execDetails[1]}\nTrace:\n{traceback.format_exc()}", quitWithStatus=1)
 
     if addToGlobal:
       globalAllTemplatesProcessed[self.name] = self
@@ -287,7 +379,7 @@ class TemplateHandler():
       self.result = self.loadedTemplate.render(values)
     except Exception as e:
       execDetails = sys.exc_info()
-      Log(f"Failed to render template at '{self.path}' with an exception from Jinja, details:\nType:{execDetails[0]}\nValue:{execDetails[1]}\nTrace:\n{traceback.format_exc()}", quitWithStatus=1)
+      Log(f"TemplateHandler.Render(): Failed to render template at '{self.path}' with an exception from Jinja, details:\nType:{execDetails[0]}\nValue:{execDetails[1]}\nTrace:\n{traceback.format_exc()}", quitWithStatus=1)
 
 def ParseValues(*AscendingPriorityInputObjects):
   if len(AscendingPriorityInputObjects) == 0:
@@ -318,41 +410,40 @@ def Main():
   # Template Handling
   for tmpl in args.templates:
     for tmplChild in tmpl:
-      fullPath = os.path.abspath(tmplChild)
-      TemplateHandler(path=fullPath, addToGlobal=True)
+      tmplFullPath = os.path.abspath(tmplChild)
+      if "*" in tmplFullPath:
+        Log(f'Main(): Handling template globbing expression {tmplFullPath}...')
+        s = tmplFullPath.find('*')
+        for globbedTemplatePath in pathlib.Path(tmplFullPath[:s]).glob(tmplFullPath[s:]):
+          if CurrentLoggingSettings.verbosity > 1:
+            Log(f'Main(): => {globbedTemplatePath}...')
+          if globbedTemplatePath.is_file():
+            TemplateHandler(path=globbedTemplatePath, addToGlobal=True)
+      else:
+        TemplateHandler(path=tmplFullPath, addToGlobal=True)
 
   ##########################################
   # Variable Handling
-  for inputsPath in args.input_values:
+  for inputsPath in args.inputs:
     for inputsPathChild in inputsPath:
-      fullPath = os.path.abspath(inputsPathChild)
-      if not os.path.exists(inputsPathChild):
-        Log(f"Could not open inputs file at path '{inputsPathChild}'", quitWithStatus=1)
-    
-      matchedType = ""
-      finalObj = None
-      with open(inputsPathChild) as f:
-        fileData = f.read()
+      inputFullPath = os.path.abspath(inputsPathChild)
+      if "*" in inputFullPath:
+        s = inputFullPath.find('*')
+        Log(f'Main(): Handling inputs globbing expression {inputFullPath}...')
+        
+        for globbedInputPath in pathlib.Path(inputFullPath[:s]).glob(inputFullPath[s:]):
+          if CurrentLoggingSettings.verbosity > 1:
+            Log(f'Main(): => {globbedInputPath}...')
+          overallValues = CombineValues(overallValues, ParseValuesFile(globbedInputPath), globbedInputPath)
+        continue
 
-      try:
-        ymlObj = yaml.load(fileData, Loader=yaml.FullLoader)
-        finalObj = ymlObj
-        matchedType = "yaml"
-      except Exception as e:
-        pass
+      if not os.path.exists(inputFullPath):
+        Log(f"Main(): Could not open inputs file at path '{inputFullPath}'", quitWithStatus=1)
 
-      if matchedType == "":
-        try:
-          jsonObj = json.loads(fileData)
-          finalObj = jsonObj
-          matchedType = "json"
-        except Exception as e:
-          pass
+      if os.stat(inputFullPath).st_size == 0:
+        continue
 
-      if matchedType == "":
-        Log(f"Could not load inputs file '{inputsPathChild}' as either a json or a yaml file, please inspect!", quitWithStatus=1)
-
-      overallValues = CombineValues(overallValues, finalObj, fullPath)
+      overallValues = CombineValues(overallValues, ParseValuesFile(inputFullPath), inputFullPath)
 
   if not args.ignore_env_vars:
     foundVars = {}
@@ -364,11 +455,26 @@ def Main():
       nestedVal = path.split(args.dict_separator)
       SetNestedValue(baseResource=overallValues, path=nestedVal, value=val)
 
+  if args.explicit:
+    foundExplicitVars = {}
+    for explicitValList in args.explicit:
+      for explicitVal in explicitValList:
+
+        splits = explicitVal.split("=")
+        foundExplicitVars[splits[0]] = "=".join(splits[1:])
+
+    for path in foundExplicitVars:
+      nestedVal = path.split(args.dict_separator)
+      SetNestedValue(baseResource=overallValues, path=nestedVal, value=foundExplicitVars[path])
+
   ##########################################
   # Templating
   for ind, tmpl in enumerate(globalAllTemplatesProcessed):
     globalAllTemplatesProcessed[tmpl].Render(overallValues)
 
+    if globalAllTemplatesProcessed[tmpl].result == "":
+      continue
+
     if args.dump_to_dir:
       dest = args.dump_to_dir + f'/{ind}-{globalAllTemplatesProcessed[tmpl].basename}'
       with open(dest, "w+") as f:
